!### macro's #####################################################
!
#define IF_ERROR_RETURN(action) if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
!
!#################################################################
module user_output_mix

  ! module included to write daily and hourly output..
  ! called from user_output, and controlg.f90 (on exit, to close files).
  ! Per day a seperate file is created:
  ! 
  ! Use for paralel: one file per PE to avoid gathering
  ! Use for dynamic timesteps: weight ndyn/ndyn_max

  use dims,  only : nregions
  use GO, only : gol, goErr

  implicit none
  ! interface

  private

  public :: accumulate_mix, write_mix
  public :: output_mix_init, output_mix_close

  ! local
  ! --- const ----------------------------------
    
  character(len=*), parameter  ::  mname = 'user_output_mix'

  type mixfile
     real,dimension(:,:,:,:),pointer         :: field  ! 3D fields to write
     real,dimension(:,:),pointer             :: ps     ! the surface pressure
     real,dimension(:,:),pointer             :: wda    ! number of averages
     character(len=20),dimension(:),pointer  :: namefield
     character(len=100) :: fname
     integer  :: funit = 0                   ! unit number 
     real     :: lon_first                   ! x geometry field
     real     :: lon_inc                     ! x geometry field
     real     :: lat_first                   ! y geometry field
     real     :: lat_inc                     ! y geometry field
     integer  :: lon_n                       ! x dimension requested field
     integer  :: lat_n                       ! y dimension requested field
     integer,dimension(6)  :: idate_start    ! date of collection!
     integer,dimension(6)  :: idate_end      ! date of collection!
  end type mixfile

  type(mixfile),dimension(nregions),save  :: mixf

  ! generate output if true:
  logical :: okdebug = .false.

contains


  subroutine output_mix_init
    !
    !
    use dims,        only : nregions, newsrun
    use dims,        only : im, jm, lm, region_name, parent,  itau
    use dims,        only : xbeg, xend, ybeg, yend, dx, dy, xref, yref
    use dims,        only : zbeg, zend, dz, zref, idatei
    use chem_param,  only : names

    use partools,    only : myid, ntracetloc, ntracet_ar

    implicit none
    ! IO
    ! local
    integer               :: region, offsetn
    integer               :: i,j,n 
    integer, dimension(6) :: idater

    offsetn = sum(ntracet_ar(0:myid-1))
    
    regionloop: do region = 1, nregions

          !  initialize the mix output: instantaneous

          allocate(mixf(region)%field(im(region),jm(region),lm(region),ntracetloc+1))  ! +1 for gph
          allocate(mixf(region)%ps(im(region),jm(region)))
          allocate(mixf(region)%wda(im(region),jm(region)))
          allocate(mixf(region)%namefield(ntracetloc+1))
          do i=1, ntracetloc
             mixf(region)%namefield(i) =  names(i+offsetn)
          enddo
          mixf(region)%namefield(ntracetloc+1) = 'gph'

          mixf(region)%lon_first = xbeg(region) + 0.5*dx/xref(region)
          mixf(region)%lat_first = ybeg(region) + 0.5*dy/yref(region)
          mixf(region)%lat_inc   = dy/yref(region)
          mixf(region)%lon_inc   = dx/xref(region)
          mixf(region)%lon_n     = im(region)
          mixf(region)%lat_n     = jm(region)
          mixf(region)%ps        = 0.0
          mixf(region)%field     = 0.0
          mixf(region)%wda       = 0.0

          mixf(region)%idate_start = idatei  ! initialize start date of interval

    end do  regionloop ! nregions

  end subroutine output_mix_init



  subroutine output_mix_close
    !-----------------------------------------------------------
    !
    !-----------------------------------------------------------
    use dims,        only : nregions

    ! local
    integer               :: region



    ! start
    do region=1,nregions

       call write_mix(region)

    enddo   !regions. Do not yet deallocate because parent mix is needed for updates!

    do region=1,nregions
       deallocate(mixf(region)%field)   ! free memory
       deallocate(mixf(region)%ps)   ! free memory
       deallocate(mixf(region)%wda)   ! free memory
       deallocate(mixf(region)%namefield)   ! free memory
    enddo   !regions


  end subroutine output_mix_close



  subroutine accumulate_mix(region)
    !-----------------------------------------------------------
    !
    !-----------------------------------------------------------
    use global_data, only : mass_dat, region_dat
    use meteo      , only : sp_dat,gph_dat
    use dims,        only : im,jm,lm, itaur, ndyn, tref, isr, jsr, ier, jer, at, bt, ndyn, ndyn_max
    use chem_param,  only : names, ra, fscale, ntracet, ntrace_chem
    use binas    ,   only : Avog,Dobs,pi, xmair
    use datetime,    only : tau2date, date2tau
    use toolbox ,    only : escape_tm
    use ParTools,    only : myid, ntracetloc, ntracet_ar

    implicit none

    ! in/out
    integer, intent(in) :: region

    ! local
    integer               :: io,istat,i,j, imr, jmr, lmr, n, l
    integer               :: sfstart, sfscatt, sfsnatt, sfend
    integer, dimension(6) :: idater
    integer               :: itau_lt, dtime, itau_out12, itau_out00, offsetn, ditau, dtime_max
    real                  :: longitude, weight

    ! MPI arrays to gather fields.
    real,dimension(:,:,:,:),               pointer     :: rm
    real,dimension(:,:,:),                 pointer     :: p
    real,dimension(:,:,:),                 pointer     :: gph
    real,dimension(:,:,:),                 pointer     :: m
    real,dimension(:),                     pointer     :: dxyp
    integer,dimension(:,:),                pointer     :: zoomed

    zoomed => region_dat(region)%zoomed

    dtime = ndyn/tref(region)   ! basic time step (seconds)
    dtime_max = ndyn_max/tref(region)   ! basic time step (seconds)
    weight = float(ndyn)/float(ndyn_max)
    offsetn = sum(ntracet_ar(0:myid-1))

    ! collect output FOR LT
    imr = im(region) ; jmr = jm(region) ; lmr = lm(region)


    m => mass_dat(region)%m_t
    rm => mass_dat(region)%rm_t
    p=> sp_dat(region)%data
    gph=> gph_dat(region)%data
    dxyp=> region_dat(region)%dxyp


!  fill fields for averaging:

    do j=jsr(region),jer(region)
       do i=isr(region), ier(region)
          if(zoomed(i,j) /= region) cycle
              ! average output over day:
              mixf(region)%wda(i,j) = mixf(region)%wda(i,j) + weight
              do l=1,lm(region)
                do n=1,ntracetloc
                   mixf(region)%field(i,j,l,n) = &
                      mixf(region)%field(i,j,l,n) + weight*rm(i,j,l,n)/m(i,j,l)*(xmair/ra(offsetn+n))
                enddo
                mixf(region)%field(i,j,l,ntracetloc+1) = &
                     mixf(region)%field(i,j,l,ntracetloc+1) + weight*gph(i,j,l)
             enddo
             mixf(region)%ps(i,j) = &
                mixf(region)%ps(i,j) + weight*p(i,j,1)
       enddo
    enddo
    nullify(rm)    ! reset pointers
    nullify(m)    ! reset pointers
    nullify(p)
    nullify(gph)
    nullify(dxyp)
    nullify(zoomed)

  end subroutine accumulate_mix

subroutine update_parent_mix(region)

  use dims
  use global_data,  only: region_dat
  use toolbox,      only : escape_tm
  use chem_param,   only: ntrace

  IMPLICIT NONE

  !__IO___________________________________________________________________

  INTEGER, INTENT(IN)   :: region

  !__LOCAL_VARIABLES______________________________________________________

  REAL, DIMENSION(:,:,:,:), POINTER    :: mean_mix       ! column value m  - mean
  REAL, DIMENSION(:,:), POINTER    :: mean_ps      ! column value ps  - mean

  REAL, DIMENSION(:,:,:,:), POINTER    :: p_mean_mix       ! column value m  - mean
  REAL, DIMENSION(:,:), POINTER    :: p_mean_ps      ! column value ps  - mean

  REAL, DIMENSION(:), POINTER    :: dxyp

  INTEGER :: i,j
  INTEGER :: ip, jp
  INTEGER :: my_parent
  INTEGER :: xref_,yref_,zref_

  INTEGER :: imp,jmp
  INTEGER :: imr,jmr

  INTEGER :: ic,jc

  INTEGER :: iox,ioy1,ioy2
  REAL    :: w,wtot



  !__START_SUBROUTINE______________________________________________________



  IF (region==1) RETURN


  imr = im(region)
  jmr = jm(region)

  ! determine parent
  my_parent = parent(region)

  xref_ = xref(region)/xref(my_parent)
  yref_ = yref(region)/yref(my_parent)

  imp = im(region)/xref_
  jmp = jm(region)/yref_

  ! check calculated imp, jmp, lmp
  IF(ibeg(region) < iend(region) .and. imp /= iend(region)-ibeg(region)+1) THEN
    call escape_tm('stopped in update_parent_columns')
  ENDIF
    
  IF(jmp.ne.jend(region)-jbeg(region)+1) THEN
    call escape_tm('stopped in update_parent_columns')
  ENDIF  
 

  mean_mix  => mixf(region)%field
  mean_ps  => mixf(region)%ps

  p_mean_mix  => mixf(my_parent)%field
  p_mean_ps => mixf(my_parent)%ps

  dxyp=> region_dat(region)%dxyp

  iox = isr(region)/xref_
  ioy1 = jsr(region)/yref_
  ioy2 = (jm(region)-jer(region)+1)/yref_


  DO jp=jbeg(region)+ioy1, jend(region)-ioy2  
    jc = (jp-jbeg(region))*yref_
    DO ip=ibeg(region)+iox, iend(region)-iox
      ic = (ip-ibeg(region))*xref_

      p_mean_ps(ip,jp) = 0.0
      p_mean_mix(ip,jp,:,:) = 0.0


      wtot = 0.0
      DO j=1,yref_
        w = dxyp(jc+j)
        DO i=1,xref_

          p_mean_mix(ip,jp,:,:) = p_mean_mix(ip,jp,:,:)  + mean_mix(ic+i,jc+j,:,:)*w
          p_mean_ps(ip,jp) = p_mean_ps(ip,jp)  + mean_ps(ic+i,jc+j)*w

          wtot = wtot+w
        ENDDO
      ENDDO
      p_mean_mix(ip,jp,:,:) = p_mean_mix(ip,jp,:,:)/wtot
      p_mean_ps(ip,jp) = p_mean_ps(ip,jp)/wtot


    ENDDO
  ENDDO


  NULLIFY(mean_mix)  
  NULLIFY(mean_ps)  

  NULLIFY(p_mean_mix)  
  NULLIFY(p_mean_ps)  

END SUBROUTINE update_parent_mix

!===========================================================================================================
!===========================================================================================================

  subroutine write_mix(region)
    !-----------------------------------------------------------
    !
    !-----------------------------------------------------------
    use global_data, only : mass_dat, region_dat
    use meteo      , only : sp_dat
    use dims,        only : im,jm,lm, itaur, ndyn, tref, isr, jsr, ier, jer, at, bt, ndyn, ndyn_max
    use dims       , only : nregions, region_name
    use dims       , only : dx, xref, xbeg, xend, ibeg, iend
    use dims       , only : dy, yref, ybeg, yend, jbeg, jend
    use dims       , only : dz, zref, zbeg, zend, lbeg, lend
    use chem_param,  only : names, ra, fscale, ntracet, ntrace_chem
    use binas    ,   only : Avog,Dobs,pi, xmair
    use io_hdf
    use file_hdf
    use datetime,    only : tau2date, date2tau
    use toolbox ,    only : escape_tm
    use ParTools,    only : myid, ntracetloc, ntracet_ar

    implicit none

    ! in/out
    integer, intent(in) :: region

    character(len=*), parameter  ::  rname = mname//'write_mix'
    ! local
    integer               :: io,istat,i,j, imr, jmr, lmr, n, l
    integer, dimension(6) :: idater
    integer               :: itau_lt, dtime, itau_out12, itau_out00, offsetn, ditau, dtime_max
    real                  :: longitude, weight

    integer,dimension(:,:),                pointer     :: zoomed
    integer :: status

    type(THdfFile)    ::  hdf
    type(TSds)        ::  sds

    zoomed => region_dat(region)%zoomed

    dtime = ndyn/tref(region)   ! basic time step (seconds)
    dtime_max = ndyn_max/tref(region)   ! basic time step (seconds)

    ! go to middle of interval 
    call tau2date(itaur(region),idater)

    mixf(region)%idate_end = idater 

    ! Create file plus header

    call mix_file_init ! open mix file

    if ( okdebug ) print *, 'output_mix: Writing daily Mix file output',region

    !WP! First divide each element by weight, then update_parents, and finally
    !write the output.

    do j=jsr(region),jer(region)
       do i=isr(region), ier(region)
          if(zoomed(i,j) /= region) cycle
          if(mixf(region)%wda(i,j)>0) &
          mixf(region)%ps(i,j)  =  mixf(region)%ps(i,j)/mixf(region)%wda(i,j)
       enddo
    enddo
    do n=1,ntracetloc+1
      do l=1,lm(region)     
         do j=jsr(region),jer(region)
            do i=isr(region),ier(region)
               if(zoomed(i,j) /= region) cycle
               if(mixf(region)%wda(i,j)>0) &
               mixf(region)%field(i,j,l,n) =  mixf(region)%field(i,j,l,n)/   &
                                                   mixf(region)%wda(i,j)
            enddo
         enddo
      enddo   
    enddo   

    !WP! Added explicit call to update_parent to ensure that inner domain of
    !parents is filled with appropriate values. This was missing before. Note
    !that the order in which to write mix files *MUST* be from smallest zoom
    !to parents.

    call update_parent_mix(region)

    call Init( sds, hdf,'ps' , (/im(region),jm(region)/), 'real(4)', status )
    call SetDim(sds,0,'LON'//region_name(region),'deg E',(/(xbeg(region)+i*dx/xref(region),i=0,im(region)-1)/),status)
    call SetDim(sds,1,'LAT'//region_name(region),'deg N',(/(ybeg(region)+i*dy/yref(region),i=0,jm(region)-1)/),status)
    call WriteData(sds,mixf(region)%ps,status)
    call WriteAttribute(sds,'idate_start',mixf(region)%idate_start,status)
    call WriteAttribute(sds,'idate_end',mixf(region)%idate_end,status)
    call WriteAttribute(sds,'units','Pa',status)
    call Done(sds,status)
    IF_ERROR_RETURN(status=1)

    do n=1,ntracetloc+1

      call Init( sds, hdf,mixf(region)%namefield(n) , (/im(region),jm(region),lm(region)/), 'real(4)', status )
      call SetDim(sds,0,'LON'//region_name(region),'deg E',(/(xbeg(region)+i*dx/xref(region),i=0,im(region)-1)/),status)
      call SetDim(sds,1,'LAT'//region_name(region),'deg N',(/(ybeg(region)+i*dy/yref(region),i=0,jm(region)-1)/),status)
      call SetDim(sds,2,'HYBRID','number',(/(i,i=1,lm(region))/),status)
      call WriteData(sds,mixf(region)%field(:,:,:,n),status)
      call WriteAttribute(sds,'idate_start',mixf(region)%idate_start,status)
      call WriteAttribute(sds,'idate_end',mixf(region)%idate_end,status)
      if(n .le. ntracetloc) then
         call WriteAttribute(sds,'units','mole/mole',status)
      else
         call WriteAttribute(sds,'units','m',status)
      endif
      call Done(sds,status)
      IF_ERROR_RETURN(status=1)
    enddo  !n

    call Done(hdf,status)
    IF_ERROR_RETURN(status=1)
    if ( okdebug ) print *, 'output_mix: Mix file lt12 close for region:', &
           region

    mixf(region)%ps        = 0.0
    mixf(region)%field     = 0.0
    mixf(region)%wda       = 0.0
    mixf(region)%idate_start = mixf(region)%idate_end ! end of interval becomes start of next interval  

    nullify(zoomed)

   contains

    subroutine mix_file_init
    use dims,        only : nregions
    use dims,        only : im, jm, lm, region_name, parent,  itaur, itau
    use dims,        only : xbeg, xend, ybeg, yend, dx, dy, xref, yref
    use dims,        only : zbeg, zend, dz, zref, at, bt
    use global_data, only : region_dat, outdir
    use datetime,    only : tau2date
    use partools,    only : myid, ntracetloc, ntracet_ar

    integer               :: sfstart, sfscatt, sfsnatt, sfend
    integer, dimension(6) :: idater
    character(len=12)      :: idates ! arj change, was len=10
    character(len=2)      :: myids
    integer		  :: io,istat

    write(myids,'(i2.2)') myid
     
    idater=mixf(region)%idate_start  ! use start of interval in file name

    write(idates, '(i4,4i2.2)') idater(1), idater(2), idater(3), idater(4), idater(5)  ! arj change, was up to idater(4)
    mixf(region)%fname = trim(outdir)//'/mix_'//region_name(region)//'_'//myids//'_'//idates//'.hdf'

    call Init( hdf, trim(mixf(region)%fname), 'create', status )
  
    if(okdebug) print *, 'init_mixdata: daily Mix file opened for region:',&
         region, hdf%id, mixf(region)%fname

    call WriteAttribute( hdf, 'region_name', region_name(region), status )
    call WriteAttribute( hdf, 'im'         , im(region)         , status )
    call WriteAttribute( hdf, 'jm'         , jm(region)         , status )
    call WriteAttribute( hdf, 'lm'         , lm(region)         , status )
    call WriteAttribute( hdf, 'dx'         , dx/xref(region)    , status )
    call WriteAttribute( hdf, 'dy'         , dy/yref(region)    , status )
    call WriteAttribute( hdf, 'dz'         , dz/zref(region)    , status )
    call WriteAttribute( hdf, 'xbeg'       , xbeg(region)       , status )
    call WriteAttribute( hdf, 'xend'       , xend(region)       , status )
    call WriteAttribute( hdf, 'ybeg'       , ybeg(region)       , status )
    call WriteAttribute( hdf, 'yend'       , yend(region)       , status )
    call WriteAttribute( hdf, 'zbeg'       , zbeg(region)       , status )
    call WriteAttribute( hdf, 'zend'       , zend(region)       , status )
    if ( region /= 1 ) then
      call WriteAttribute( hdf, 'ibeg', ibeg(region), status )
      call WriteAttribute( hdf, 'iend', iend(region), status )
      call WriteAttribute( hdf, 'jbeg', jbeg(region), status )
      call WriteAttribute( hdf, 'jend', jend(region), status )
      call WriteAttribute( hdf, 'lbeg', lbeg(region), status )
      call WriteAttribute( hdf, 'lend', lend(region), status )
    end if

    call WriteAttribute( hdf, 'xref'       , xref(region), status )
    call WriteAttribute( hdf, 'yref'       , yref(region), status )
    call WriteAttribute( hdf, 'zref'       , zref(region), status )
    call WriteAttribute( hdf, 'tref'       , tref(region), status )
    call WriteAttribute( hdf, 'at'         , at          , status )
    call WriteAttribute( hdf, 'bt'         , bt          , status )
    call WriteAttribute( hdf, 'idate_start'         , mixf(region)%idate_start          , status )
    call WriteAttribute( hdf, 'idate_end'         , mixf(region)%idate_end          , status )

    end subroutine mix_file_init

  end subroutine write_mix
  
end module user_output_mix
